import os
import re
import unicodedata
from datetime import datetime

from kanjize import kanji2int

KANSUJI = '〇一二三四五六七八九'
KANSUJI_DIGIT = '十百千万'


def get_env_or_default(key, default):
    return os.environ[key] if key in os.environ else default


def to_date_str(dt):
    return '{:02d}-{:02d}-{:02d}'.format(dt.year, dt.month, dt.day)


def filter_dict_by_value(dict_, num_items):
    """
    valueが大きい上位のitemに絞る
    """

    sorted_items = sorted(dict_.items(), key=lambda x: x[1], reverse=True)
    num_items = min(len(dict_), num_items)
    return dict(sorted_items[:num_items])


def filter_dict_by_thresh(dict_, thresh):
    """
    valueがthresh以上のitemに絞る
    """

    return dict(filter(lambda x: x[1] >= thresh, dict_.items()))


def contains_word(text, allow_list=None, block_list=None):
    """
    単語を含むか判定する
    allowとblockの両方が含まれていた場合はFalseを返す
    """

    allow_list = allow_list or list()
    block_list = block_list or list()
    has_allow_word = any(w in text for w in allow_list)
    has_block_word = any(w in text for w in block_list)
    return has_allow_word and not has_block_word


def strip_join(str_list, sep=''):
    return sep.join(map(lambda x: x.strip(), str_list))


def deduplicate(items):
    """
    順番を保持したままリストの重複を除く
    """

    return list(dict.fromkeys(items))


def convert_kansuji_to_int(s):
    buffer = ''
    sr = s[::-1]  # digitを後ろから変換する
    for i, si in enumerate(sr):
        if si in KANSUJI_DIGIT:
            zeros = KANSUJI_DIGIT.find(si) + 1
            if len(buffer) > zeros:
                raise ValueError(f'too many numbers before "{si}": {s}')
            buffer += '〇' * (zeros - len(buffer))  # 〇で桁を埋める
            if i == len(sr) - 1 or sr[i + 1] not in KANSUJI:
                buffer += '一'
        elif si in KANSUJI:
            buffer += si
        else:
            raise ValueError(f'found non-kansuji "{si}": {s}')
    return int(''.join([str(int(unicodedata.numeric(x))) for x in buffer[::-1]]))


def replace_kansuji_to_number(s):
    pattern = f'[{KANSUJI + KANSUJI_DIGIT}]+'
    return re.sub(pattern, lambda x: str(convert_kansuji_to_int(x.group(0))), s)


def get_str_offset(s):
    for i, c in enumerate(s):
        if c not in {' ', '　'}:
            return i
    return -1


class DateConverter:
    era2year = {'明治': 1872, '大正': 1912, '昭和': 1925, '平成': 1989, '令和': 2019}
    pattern = r'((?:明治|大正|昭和|平成|令和).*)年(.*)月(.*)日'

    @staticmethod
    def convert(s):
        """
        和暦の日付をdatetimeに変換する
        """

        m = re.search(DateConverter.pattern, s)
        if not m:
            raise ValueError(f'{s} does not fully match with {DateConverter.pattern}')
        year = DateConverter.convert_year_to_int(m.group(1))
        month = DateConverter.convert_to_int(m.group(2))
        day = DateConverter.convert_to_int(m.group(3))
        return datetime(year=year, month=month, day=day)

    @staticmethod
    def convert_year_to_int(s):
        s = s.strip()
        for era, year in DateConverter.era2year.items():
            if s.startswith(era):
                offset = DateConverter.convert_to_int(s[len(era):]) - 1
                return year + offset
        return DateConverter.convert_to_int(s)

    @staticmethod
    def convert_to_int(s):
        s = s.strip()
        if s.isdigit():
            return int(s)
        elif s == '元':
            return 1
        else:
            return kanji2int(s)



from enum import Enum

import math
from elasticsearch import Elasticsearch, helpers
from elasticsearch_dsl import Search

from politylink.elasticsearch import ELASTICSEARCH_URL
from politylink.elasticsearch.schema import AbstractText, to_cls


class OpType(str, Enum):
    INDEX = 'index'  # create or overwrite
    UPDATE = 'update'  # NotFoundError or update
    MERGE = 'merge'  # create or update (round-trip)


class ElasticsearchException(Exception):
    pass


class ElasticsearchClient:
    """
    Elasticsearch client for politylink endpoint
    """

    def __init__(self, url=ELASTICSEARCH_URL):
        self.client = Elasticsearch(url)

    def index(self, obj, op_type=OpType.INDEX):
        """
        create or update a document
        """

        assert isinstance(obj, AbstractText)
        try:
            if op_type == OpType.INDEX:
                return self.client.index(index=obj.index, id=obj.id, body=obj.__dict__)
            elif op_type == OpType.UPDATE:
                return self.client.update(index=obj.index, id=obj.id, body={'doc': obj.__dict__})
            elif op_type == OpType.MERGE:
                if self.exists(obj.id):
                    self.index(obj, op_type=OpType.UPDATE)
                else:
                    self.index(obj, op_type=OpType.INDEX)
            else:
                raise ElasticsearchException(f'unknown index operation type: {op_type}')

        except Exception as e:
            raise ElasticsearchException(f'failed to index {obj}') from e

    def bulk_index(self, objects, op_type=OpType.INDEX):
        def actions():
            for obj in objects:
                assert isinstance(obj, AbstractText)
                if op_type == OpType.INDEX:
                    yield {'_index': obj.index, '_id': obj.id, '_source': obj.__dict__, '_op_type': op_type.value}
                elif op_type == OpType.UPDATE:
                    yield {'_index': obj.index, '_id': obj.id, 'doc': obj.__dict__, '_op_type': op_type.value}
                else:
                    raise ElasticsearchException(f'unknown index operation type: {op_type}')

        try:
            return helpers.bulk(client=self.client, actions=actions())
        except Exception as e:
            raise ElasticsearchException(f'failed to bulk index') from e

    def get(self, id_):
        """
        get a document by politylink id (ref idgen)
        """

        try:
            cls = to_cls(id_)
            res = self.client.get(index=cls.index, id=id_)
            return cls(res['_source'])
        except Exception as e:
            raise ElasticsearchException(f'failed to get {id_}') from e

    def exists(self, id_):
        try:
            cls = to_cls(id_)
            return self.client.exists(index=cls.index, id=id_)
        except Exception as e:
            raise ElasticsearchException(f'failed to check existence of {id_}') from e

    def search(self, cls: AbstractText, query: str = None, start_date_str: str = None, end_date_str: str = None):
        s = Search(using=self.client, index=cls.index)
        if query:
            s = s.query('multi_match', query=query, fields=cls.get_all_fields())
        if start_date_str:
            s = s.filter('range', **{cls.Field.DATE: {'gte': start_date_str}})
        if end_date_str:
            s = s.filter('range', **{cls.Field.DATE: {'lt': end_date_str}})
        try:
            res = s.execute()
            return list(map(lambda hit: cls(hit['_source']), res['hits']['hits']))
        except Exception as e:
            raise ElasticsearchException(f'failed to search {cls.__class__.__name__} for {s.to_dict()}') from e

    def get_term_statistics(self, id_):
        """
        get term statistics using termvectors API

        :param id_: politylink ID
        :return: dictionary of term to statistics (tf, idf, tfidf)
        """

        try:
            cls = to_cls(id_)
            res = self.client.termvectors(index=cls.index, id=id_, params={'fields': 'body', 'term_statistics': 'true'})
        except Exception as e:
            raise ElasticsearchException(f'failed to get termvectors for {id_}') from e

        term2stats = dict()
        doc_count = res['term_vectors']['body']['field_statistics']['doc_count']
        for term, stats_raw in res['term_vectors']['body']['terms'].items():
            stats = dict()
            stats['tf'] = stats_raw['term_freq']
            stats['idf'] = math.log(doc_count / stats_raw['doc_freq'])
            stats['tfidf'] = stats['tf'] * stats['idf']
            term2stats[term] = stats
        return term2stats



from enum import Enum

import politylink.graphql.schema
from politylink.graphql.schema import Bill


def to_cls(id_):
    """
    map Politylink ID to ElasticSearch schema class
    """

    if id_.startswith('News'):
        return NewsText
    elif id_.startswith('Bill'):
        return BillText
    elif id_.startswith('Minutes'):
        return MinutesText
    elif id_.startswith('Speech'):
        return SpeechText
    else:
        raise ValueError(f'unknown id type: {id_}')


class AbstractText:
    """
    Abstract class to define elasticsearch document schema
    """

    # define index name here
    index = NotImplemented

    # define document schema here
    class Field(str, Enum):
        ID = 'id'  # 'id' is required
        DATE = 'date'  # 2020-01-01

    def __init__(self, dct=None):
        if dct:
            for field in self.Field:
                key = field.value
                if key in dct:
                    setattr(self, key, dct[key])

    def set(self, field: Field, value):
        setattr(self, field.value, value)

    def __repr__(self):
        return f'<{self.__class__.__name__}: {self.__dict__}>'

    @classmethod
    def get_all_fields(cls):
        """
        return all field names except for ID and DATE
        """

        fields = []
        for field in cls.Field:
            if field.name not in {cls.Field.ID.name, cls.Field.DATE.name}:
                fields.append(field.value)
        return fields


class NewsText(AbstractText):
    index = 'news'

    class Field(str, Enum):
        ID = 'id'
        TITLE = 'title'
        BODY = 'body'
        DATE = 'date'


class BillText(AbstractText):
    index = 'bill'

    class Field(str, Enum):
        ID = 'id'
        BILL_NUMBER = 'bill_number'
        TITLE = 'title'
        BODY = 'body'
        SUPPLEMENT = 'supplement'
        REASON = 'reason'
        ALIASES = 'aliases'
        TAGS = 'tags'
        CATEGORY = 'category'
        STATUS = 'status'
        SUBMITTED_DATE = 'submitted_date'
        LAST_UPDATED_DATE = 'last_updated_date'
        GROUPS = 'groups'
        MINISTRIES = 'ministries'
        SUBMITTED_DIET = 'submitted_diet'
        BELONGED_TO_DIETS = 'belonged_to_diets'
        SUBMITTED_GROUPS = 'submitted_groups'
        SUPPORTED_GROUPS = 'supported_groups'
        OPPOSED_GROUPS = 'opposed_groups'


class MinutesText(AbstractText):
    index = 'minutes'

    class Field(str, Enum):
        ID = 'id'
        TITLE = 'title'
        BODY = 'body'
        DATE = 'date'


class SpeechText(AbstractText):
    index = 'speech'

    class Field(str, Enum):
        ID = 'id'
        TITLE = 'title'
        SPEAKER = 'speaker'
        BODY = 'body'
        DATE = 'date'


class IndexedEnum(Enum):
    def __init__(self, index, *args, **kwargs):
        self.index = index

    @classmethod
    def from_index(cls, index: int):
        for e in cls:
            if e.index == index:
                return e
        raise ValueError(f'index not found: {index}')


class BillStatus(IndexedEnum):
    SUBMITTED = (0, '提出')
    PASSED_REPRESENTATIVES_COMMITTEE = (1, '衆委可決')
    PASSED_REPRESENTATIVES = (2, '衆可決')
    PASSED_COUNCILORS_COMMITTEE = (3, '参委可決')
    PASSED_COUNCILORS = (4, '参可決')
    PROCLAIMED = (5, '公布')

    def __init__(self, index, label):
        super().__init__(index)
        self.label = label

    @staticmethod
    def from_gql(bill: Bill):
        max_date, max_bill_status = '', None
        for bill_status in BillStatus:
            gql_date_field = bill_status.name.lower() + '_date'
            if hasattr(bill, gql_date_field):
                date = getattr(bill, gql_date_field).formatted
                if date and date >= max_date:  # >= to prioritize later
                    max_date = date
                    max_bill_status = bill_status
        if max_bill_status is None:
            raise ValueError(f'bill does not have valid date: {bill}')
        return max_bill_status


class BillCategory(IndexedEnum):
    KAKUHOU = (0, '閣法')
    SHUHOU = (1, '衆法')
    SANHOU = (2, '参法')

    def __init__(self, index, label):
        super().__init__(index)
        self.label = label

    @staticmethod
    def from_gql(bill: Bill):
        for bill_category in BillCategory:
            if bill.category == bill_category.name:
                return bill_category
        raise ValueError(f'bill does not have valid category: {bill.category}')


class ParliamentaryGroup(IndexedEnum):
    JIMIN = (0, '自民')
    RIKKEN = (1, '立憲')
    KOMEI = (2, '公明')
    KYOSAN = (3, '共産')
    ISHIN = (4, '維新')
    KOKUMIN = (5, '国民')

    def __init__(self, index, label):
        super().__init__(index)
        self.label = label

    @staticmethod
    def from_gql(gql_group: politylink.graphql.schema.ParliamentaryGroup):
        for es_group in ParliamentaryGroup:
            if gql_group == es_group.name:
                return es_group
        raise ValueError(f'failed to map ParliamentaryGroup: {gql_group}')



from os import path

import spacy
from spacy.lang.ja import stop_words

nlp = spacy.load('ja_ginza')

WORDCLOUD_POS_TAGS = ['ADJ', 'ADV', 'NOUN', 'PROPN']
STOPWORDS_SLOTHLIB_PATH = path.join(path.dirname(__file__), 'stopwords_slothlib.txt')
STOPWORDS_MINUTES_PATH = path.join(path.dirname(__file__), 'stopwords_minutes.txt')


def filter_by_pos(texts, pos_tags):
    """
    filter text that includes target Part-Of-Speech tags
    https://spacy.io/api/annotation#pos-tagging
    """

    ret = []
    for doc in nlp.pipe(texts):
        for token in doc:
            if token.pos_ in pos_tags:
                ret.append(doc.text)
                break
    return ret


def load_stopwords(file_path):
    with open(file_path, mode='r') as f:
        stopwords = f.read().splitlines()
    return set(stopwords)


STOPWORDS = set()
STOPWORDS.update(load_stopwords(STOPWORDS_SLOTHLIB_PATH))
STOPWORDS.update(stop_words.STOP_WORDS)



import re
from logging import getLogger
from typing import List

from politylink.graphql.schema import BillActionType, BillCategory
from politylink.utils import contains_word, replace_kansuji_to_number

LOGGER = getLogger(__name__)


def extract_bill_number_or_none(text, short=False):
    pattern = r'第?([0-9]+)回?-?(国会)?(閣|衆|参)法?-?第?([0-9]+)号?'
    match = re.search(pattern, replace_kansuji_to_number(text))
    if not match:
        return None
    template = '{}-{}-{}' if short else '第{}回国会{}法第{}号'
    return template.format(match.group(1), match.group(3), match.group(4))


def extract_bill_category_or_none(text) -> BillCategory:
    if contains_word(text, ['内閣提出', '閣法']):
        return BillCategory.KAKUHOU
    elif contains_word(text, ['衆議院提出', '衆法']):
        return BillCategory.SHUHOU
    elif contains_word(text, ['参議院提出', '参法']):
        return BillCategory.SANHOU
    return None


def extract_bill_action_types(speech) -> List[BillActionType]:
    action_lst = []
    if contains_word(speech, ['説明'], ['省略', '終わり', '既に聴取']):
        if contains_word(speech, ['修正案']):
            action_lst.append(BillActionType.AMENDMENT_EXPLANATION)
        elif contains_word(speech, ['附帯決議']):
            action_lst.append(BillActionType.SUPPLEMENTARY_EXPLANATION)
        elif contains_word(speech, ['趣旨の説明', '趣旨説明']):
            action_lst.append(BillActionType.BILL_EXPLANATION)
    if contains_word(speech, ['質疑']):
        action_lst.append(BillActionType.QUESTION)
    if contains_word(speech, ['討論']):
        action_lst.append(BillActionType.DEBATE)
    if contains_word(speech, ['採決']):
        action_lst.append(BillActionType.VOTE)
    if contains_word(speech, ['委員長の報告']):
        action_lst.append(BillActionType.REPORT)
    return action_lst
